//	Zinc Application Framework - Z_UTIME1.CPP
//	COPYRIGHT (C) 1990-1998.  All Rights Reserved.
//	Zinc Software Incorporated.  Pleasant Grove, Utah  USA

#include <z_utime1.hpp>
#include <z_string.hpp>
#include <z_error.hpp>
#include <z_app.hpp>
#define ZAF_UTIME_INFO
#include "lang_def.cpp"
#include "gbl_def.cpp"

static ZafIChar utimeSeparator[3] = { '.','.',0 };
static ZafIChar rangeSeparator[3] = { ',',' ',0 };

// ----- ZafUTime -----------------------------------------------------------

ZafUTime::ZafUTime(int left, int top, int width, ZafUTimeData *zUTimeData) :
	ZafString(left, top, width, new ZafStringData),
	utimeData(ZAF_NULLP(ZafUTimeData))
{
	// Get the messages.
	messages = ZafLanguageManager::Allocate(className, defaultMessages);

	// Initialize the utime data.
	SetUTimeData(zUTimeData ? zUTimeData : new ZafUTimeData);
}

ZafUTime::ZafUTime(int left, int top, int width, int year, int month,
	int day, int hour, int minute, int second, int milliSecond) :
	ZafString(left, top, width, new ZafStringData),
	utimeData(ZAF_NULLP(ZafUTimeData))
{
	// Get the messages.
	messages = ZafLanguageManager::Allocate(className, defaultMessages);

	// Initialize the utime data.
	SetUTimeData(new ZafUTimeData(year, month, day, hour, minute, second, milliSecond));
}

ZafUTime::ZafUTime(const ZafUTime &copy) :
	ZafString(copy), utimeData(ZAF_NULLP(ZafUTimeData))
{
	// Get the messages.
	messages = ZafLanguageManager::Allocate(className, defaultMessages);

	// Copy the time data.
	if (copy.UTimeData()->Destroyable())
		SetUTimeData(new ZafUTimeData(*copy.UTimeData()));
	else
		SetUTimeData(copy.UTimeData());
}

ZafUTime::~ZafUTime(void)
{
	// Remove the data notification.
	utimeData->SubtractNotification(this, (ZafUpdateFunction)Update);

	// Free the messages.
	ZafLanguageManager::Free(messages);

	// Restore the time information.
	if (utimeData->Destroyable() && utimeData->NotifyCount() == 0)
		delete utimeData;
}

ZafError ZafUTime::Decrement(ZafData *data, bool wrapData)
{
	// Check for a valid object.
	ZafUTimeData *utime = DynamicPtrCast(data, ZafUTimeData);
	if (!utime)
		return (ZAF_ERROR_INVALID_CLASS);

	// Decrement the value.
	ZafUTimeData _utime(*utimeData);
	OSGetUTime();
	utimeData->Decrement(utime);	// update is automatic
	if (Validate(false) != ZAF_ERROR_NONE)
	{
		const ZafIChar *rangeText = RangeText();
		ZafIChar minTime[ZAF_STRING_DATA_LENGTH];
		ZafIChar maxTime[ZAF_STRING_DATA_LENGTH];
		if (wrapData && rangeText)
		{
			ParseRange(rangeText, minTime, maxTime);
			if (*maxTime)
				utimeData->SetUTime(maxTime,
					ZafLocaleData::canonicalLocale->dateTimeStringInputFormat);
			else
				utimeData->SetUTime(_utime); // update is automatic.

		}
		else
			utimeData->SetUTime(_utime); // update is automatic.
		SetError(ZAF_ERROR_NONE);
	}

	// Return success.
	return (ZAF_ERROR_NONE);
}

ZafWindowObject *ZafUTime::Duplicate(void)
{
	return (new ZafUTime(*this));
}

ZafEventType ZafUTime::Event(const ZafEventStruct &event)
{
	// Process OS specific messages.
	ZafEventType ccode = event.type;
	if (event.InputType() == E_KEY)
		ccode = LogicalEvent(event);

	// Check for zinc events.
	switch (ccode)
	{
	// ----- Create messages -----
	case S_INITIALIZE:
		OSSetUTime();
		ccode = ZafString::Event(event);
		break;

	// ----- Action messages -----
	case L_SELECT:
	case N_NON_CURRENT:
		if (!error)
		{
			// Errors are reported only when focus changes within a window.
			bool leavingWindow = (ccode == N_NON_CURRENT && (!event.windowObject ||
				event.windowObject->RootObject() != RootObject())) ? true : false;

			// Keep a temporary time in case of error.
			ZafUTimeData oldValue(*utimeData);
			ZafError preValidateError = OSGetUTime();


			// Report the error if not leaving the window.
			if (preValidateError != ZAF_ERROR_NONE && !leavingWindow)
			if (error != ZAF_ERROR_NONE)
			{
				ZafIChar *title = ZafLanguageManager::blankString;
				ZafIChar *errorString = messages->GetMessage(ZAF_ERROR_INVALID, true);
				zafErrorSystem->ReportError(windowManager, title, ZAF_DIALOG_OK, errorString, stringData->Text());
			}

			// See if the data has changed.
			newData = (*utimeData != oldValue || ((stringData->Length() == 0) != (unanswered == true))) ? true : false;

			// Call the user/validate function.
			ccode = ZafString::Event(event);

			// Check for errors.
			if (preValidateError != ZAF_ERROR_NONE)
				ccode = leavingWindow ? ZAF_ERROR_LEAVE_INVALID : preValidateError;
			if (ccode != 0)
			{
				if (ccode != ZAF_ERROR_LEAVE_INVALID)
					utimeData->SetUTime(oldValue);
				else if (leavingWindow)
				{
					utimeData->SubtractNotification(this, (ZafUpdateFunction)Update);
					utimeData->SetUTime(oldValue);
					utimeData->AddNotification(this, (ZafUpdateFunction)Update);
				}
			}
			else if (!unanswered)
				OSSetUTime();

			// Clear the error code.
			SetError(ZAF_ERROR_NONE);
		}
		else
			// Nested call.
			return (ZAF_ERROR_LEAVE_INVALID);
		break;

	case S_COPY_DATA:
		{
		ZafWindowObject *object = event.windowObject;
		ZafUTime *utime = DynamicPtrCast(object, ZafUTime);
		if (utime)
			utimeData->SetUTime(*utime->UTimeData());
		}
		break;

	case S_SET_DATA:
		if (event.windowObject)
		{
			ZafWindowObject *object = event.windowObject;
			ZafUTime *utime = DynamicPtrCast(object, ZafUTime);
			if (utime)
				SetUTimeData(utime->UTimeData());
		}
		else
			SetUTimeData(new ZafUTimeData(*utimeData));
		break;

	case N_RESET_I18N:
		OSSetUTime();
		break;

	// ----- Default or unknown messages -----
	default:
		ccode = ZafString::Event(event);
		break;
	}

	// Return the control code.
	return (ccode);
}

ZafError ZafUTime::Increment(ZafData *data, bool wrapData)
{
	// Check for a valid object.
	ZafUTimeData *utime = DynamicPtrCast(data, ZafUTimeData);
	if (!utime)
		return (ZAF_ERROR_INVALID_CLASS);

	// Increment the value.
	ZafUTimeData _utime(*utimeData);
	OSGetUTime();
	utimeData->Increment(utime);	// update is automatic
	if (Validate(false) != ZAF_ERROR_NONE)
	{
		const ZafIChar *rangeText = RangeText();
		ZafIChar minTime[ZAF_STRING_DATA_LENGTH];
		ZafIChar maxTime[ZAF_STRING_DATA_LENGTH];
		if (wrapData && rangeText)
		{
			ParseRange(rangeText, minTime, maxTime);
			if (*minTime)
				utimeData->SetUTime(minTime,
					ZafLocaleData::canonicalLocale->dateTimeStringInputFormat);
			else
				utimeData->SetUTime(_utime); // update is automatic.

		}
		else
			utimeData->SetUTime(_utime); // update is automatic.
		SetError(ZAF_ERROR_NONE);
	}

	// Return success.
	return (ZAF_ERROR_NONE);
}

bool ZafUTime::IsA(ZafClassID compareID) const
{
	return ((compareID == classID) ? true : ZafString::IsA(compareID));
}

bool ZafUTime::IsA(ZafClassName compareName) const
{
	return (!strcmp(compareName, className) ? true : ZafString::IsA(compareName));
}

ZafError ZafUTime::Validate(bool processError)
{
	// Check for an absolute time error. Don't set the error member to ZAF_ERROR_NONE.
	ZafUTimeData currentTime;
	ZafIChar *stringTime = stringData->DynamicText();
	error = currentTime.SetUTime(stringTime, InputFormatText());

	// Check for a range error.
	if (RangeText() && error == ZAF_ERROR_NONE)
		error = ZAF_ERROR_OUT_OF_RANGE;
	int numRanges = 0;
	ZafUTimeData low, high;
	ZafIChar minTime[ZAF_STRING_DATA_LENGTH];
	ZafIChar maxTime[ZAF_STRING_DATA_LENGTH];
	ZafIChar newTime[ZAF_STRING_DATA_LENGTH];
	ZafIChar rBuffer[1024]; // Localized range string for error messages.
	rBuffer[0] = '\0';
	for (const ZafIChar *tRange = RangeText(); tRange && error == ZAF_ERROR_OUT_OF_RANGE; numRanges++)
	{
		tRange = ParseRange(tRange, minTime, maxTime);
		ZafLocaleData *temp = zafLocale;
		zafLocale = ZafLocaleData::canonicalLocale;
		low.SetUTime(minTime, ZafLocaleData::canonicalLocale->dateTimeStringInputFormat);
		high.SetUTime(maxTime, ZafLocaleData::canonicalLocale->dateTimeStringInputFormat);
		zafLocale = temp;
		if ((!minTime[0] || currentTime >= low) && (!maxTime[0] || currentTime <= high))
			error = ZAF_ERROR_NONE;
		else if (processError)
		{
			if (rBuffer[0])
				strcat(rBuffer, rangeSeparator);
			if (minTime[0])
				low.FormattedText(&rBuffer[strlen(rBuffer)], ZAF_STRING_DATA_LENGTH, OutputFormatText());
			strcat(rBuffer, utimeSeparator);
			if (maxTime[0])
				high.FormattedText(&rBuffer[strlen(rBuffer)], ZAF_STRING_DATA_LENGTH, OutputFormatText());
		}
	}

	// Process the error code.
	SetInvalid(false);
	if (error == ZAF_ERROR_NONE)
	{
		delete []stringTime;
		return (error);
	}
	else
	{
		// Keep the error code.
		if (!zafErrorSystem || !processError || !numRanges)
		{
			// Restore the original time.
			utimeData->FormattedText(newTime, ZAF_STRING_DATA_LENGTH, OutputFormatText());
			stringData->SetText(newTime);
			delete []stringTime;
			return (error);
		}
	}

	// Check for open-ended range errors.
	if (error == ZAF_ERROR_OUT_OF_RANGE && numRanges == 1)
	{
		if (minTime[0] && !maxTime[0])
		{
			error = ZAF_ERROR_LESS_THAN_RANGE;
			low.FormattedText(rBuffer, ZAF_STRING_DATA_LENGTH, OutputFormatText());
		}
		else if (!minTime[0] && maxTime[0])
		{
			error = ZAF_ERROR_GREATER_THAN_RANGE;
			high.FormattedText(rBuffer, ZAF_STRING_DATA_LENGTH, OutputFormatText());
		}
	}

	// Generate the error message and wait for a response.
	ZafIChar *title = ZafLanguageManager::blankString;
	ZafIChar *errorString = messages->GetMessage((ZafNumberID)error, true);
	if (errorString &&
		zafErrorSystem->ReportError(windowManager, title, ZAF_DIALOG_OK | ZAF_DIALOG_CANCEL, errorString, stringTime, rBuffer) == S_DLG_OK)
		error = ZAF_ERROR_LEAVE_INVALID; // Keep the new value.
	delete []stringTime;
	return (error);
}

ZafError ZafUTime::SetUTimeData(ZafUTimeData *tUTimeData)
{
	// Remove the data notification.
	if (utimeData)
		utimeData->SubtractNotification(this, (ZafUpdateFunction)Update);

	// Reset the utime data.
	if (utimeData && utimeData != tUTimeData &&
		utimeData->Destroyable() && utimeData->NotifyCount() == 0)
		delete utimeData;
	utimeData = tUTimeData ? tUTimeData : new ZafUTimeData;

	// Add the data notification.
	utimeData->AddNotification(this, (ZafUpdateFunction)Update);
	return (OSSetUTime());
}

bool ZafUTime::SetLowerCase(bool )
{
	// lowerCase is set to false by this class.
	lowerCase = false;
	return (lowerCase);
}

bool ZafUTime::SetPassword(bool )
{
	// password is set to false by this class.
	password = false;
	return (password);
}

ZafError ZafUTime::SetStringData(ZafStringData *)
{
	// String data cannot be set for this class.
	return (ZAF_ERROR_INVALID);
}

bool ZafUTime::SetUpperCase(bool )
{
	// upperCase is set to false by this class.
	upperCase = false;
	return (upperCase);
}

bool ZafUTime::SetVariableName(bool )
{
	// variableName is false for this class.
	variableName = false;
	return (variableName);
}

ZafError ZafUTime::OSGetUTime(void)
{
	// Disable notification.
	ZafUpdateType update = utimeData->Update(this);
	utimeData->SetUpdate(this, ZAF_UPDATE_NONE);

	// Set the utime based on the string data.
	OSGetText();
	error = utimeData->SetUTime(stringData->Text(), InputFormatText());

	// Restore notification.
	utimeData->SetUpdate(this, update);
	return (error);
}

ZafError ZafUTime::OSSetUTime(void)
{
	if (unanswered)
		stringData->Clear();
	else
	{
		// Set the string based on the utime data.
		ZafIChar text[ZAF_STRING_DATA_LENGTH];
		utimeData->FormattedText(text, ZAF_STRING_DATA_LENGTH, OutputFormatText());
		stringData->SetText(text);
	}

	// Return the current status.
	return (ZAF_ERROR_NONE);
}

// ----- Persistent functions -----------------------------------------------

ZafUTime::ZafUTime(const ZafIChar *name, ZafObjectPersistence &persist) :
	ZafString(name, persist.PushLevel(className, classID, ZAF_PERSIST_DIRECTORY)),
	utimeData(ZAF_NULLP(ZafUTimeData))
{
	// Get the messages.
	messages = ZafLanguageManager::Allocate(className, defaultMessages);

	if (persist.Error() == ZAF_ERROR_NONE)
		ZafUTime::ReadData(persist);
	persist.PopLevel();
	if (persist.Error() != ZAF_ERROR_NONE)
		SetError(persist.Error());
}

ZafElement *ZafUTime::Read(const ZafIChar *name, ZafObjectPersistence &persist)
{
	return (new ZafUTime(name, persist));
}

void ZafUTime::ReadData(ZafObjectPersistence &persist)
{
	// Read the data.
	ZafIChar utimeName[ZAF_MAXNAMELEN];
	ZafFile *file = persist.CurrentFile();
	*file >> utimeName;
	SetUTimeData(utimeName[0] ? new ZafUTimeData(utimeName, persist) : new ZafUTimeData);
}

void ZafUTime::Write(ZafObjectPersistence &persist)
{
	// Write the object.
	ZafString::Write(persist.PushLevel(className, classID, ZAF_PERSIST_DIRECTORY));
	if (persist.Error() == ZAF_ERROR_NONE)
		ZafUTime::WriteData(persist);
	persist.PopLevel();
	if (persist.Error() != ZAF_ERROR_NONE)
		SetError(persist.Error());
}

void ZafUTime::WriteData(ZafObjectPersistence &persist)
{
	// Write the data.
	const ZafIChar *utimeName = utimeData ? utimeData->StringID() : ZAF_NULLP(ZafIChar);
	ZafFile *file = persist.CurrentFile();
	*file << utimeName;
	if (utimeData)
		utimeData->Write(persist);
}

